"""Utility functions for accessing the SQL database"""

from __future__ import annotations

__all__ = (
    'get_job', 'get_job_status', 'get_logs', 'get_progress', 'insert_job',
    'insert_logs', 'update_job', 'upsert_progress', 'CURRENT_TIMESTAMP'
)

from typing import TYPE_CHECKING

import sqlalchemy as sql
from sqlalchemy.dialects.postgresql import insert as postgres_insert

from . import database
from .error import JobNotFoundError
from .util import Job, Log, Progress, StatusEnum

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable, Sequence
    from typing import Any, Final
    from uuid import UUID

_db = database.db

CURRENT_TIMESTAMP: Final = sql.func.current_timestamp()
"""SQL function that represents the current date and time

This can be used as a column value for `update_job()`
"""


def get_job(job_id: UUID) -> Job:
    """Returns the job that matches the given job ID

    If no job is found, it raises `JobNotFoundError`
    """
    result = _db.session.get(database.Job, job_id)
    if result is None:
        raise JobNotFoundError(job_id)

    job = Job(
        id=result.id,
        status=result.status,
        error=result.error,
        urls=result.urls,
        ytdl_opts=result.ytdl_opts,
        started=result.started,
        finished=result.finished
    )
    return job


def get_job_status(job_id: UUID) -> StatusEnum:
    """Returns just the status of the job that matches the given job ID

    If no job is found, it raises `JobNotFoundError`
    """
    statement = sql.select(database.Job.status).filter(
        database.Job.id == job_id
    )
    result = _db.session.execute(statement).one_or_none()

    if result is None:
        raise JobNotFoundError(job_id)
    else:
        status: StatusEnum = result.status
        return status


def get_logs(job_id: UUID) -> tuple[Log, ...]:
    statement = sql.select(
        database.Log.level, database.Log.message, database.Log.timestamp
    ).where(
        database.Log.job_id == job_id
    ).order_by(database.Log.timestamp)
    result = _db.session.execute(statement)

    logs = []
    for log in result:
        logs.append(Log(
            level=log.level,
            message=log.message,
            timestamp=log.timestamp
        ))
    return tuple(logs)


def get_progress(job_id: UUID) -> tuple[Progress, ...]:
    statement = sql.select(
        database.Progress.progress, database.Progress.timestamp
    ).where(
        database.Progress.job_id == job_id
    ).order_by(database.Progress.timestamp)
    result = _db.session.execute(statement)

    progress_list = []
    for progress in result:
        progress_list.append(Progress(
            progress=progress.progress,
            timestamp=progress.timestamp
        ))
    return tuple(progress_list)


def _execute(
    statement: sql.sql.expression.Executable, commit: bool
) -> sql.engine.Result:
    """Execute the given SQL statement and return the result

    If `commit` is `True`, the statement will be run within a new
    transaction, which is then commited.
    """
    if commit:
        with _db.session.begin():
            result = _db.session.execute(statement)
    else:
        result = _db.session.execute(statement)

    return result


# Each instance should be a subclass of `_db.Model`. `_db` is
# constructed using an untyped metaclass, so type hints don't work.
def _add_all(instances: Iterable[Any], commit: bool) -> None:
    """Insert the given database objects into the session

    If `commit` is `True`, the statement will be run within a new
    transaction, which is then commited.
    """
    if commit:
        with _db.session.begin():
            _db.session.add_all(instances)
    else:
        _db.session.add_all(instances)


def update_job(job_id: UUID, *, commit: bool = True, **columns: Any) -> None:
    """Update all given columns for the job that matches the job ID

    If `commit` is `True`, the statement will be run within a new
    transaction, which is then commited. Set this to `False` if you want
    to manage your own transactions.
    """
    statement = sql.update(database.Job).where(
        database.Job.id == job_id
    ).values(**columns)

    _execute(statement, commit)


def insert_job(urls: Sequence[str], *, commit: bool = True) -> UUID:
    """Insert a new job and return the job ID

    This must be called before starting the download via `ytdl`

    If `commit` is `True`, the statement will be run within a new
    transaction, which is then commited. Set this to `False` if you want
    to manage your own transactions.
    """
    statement = sql.insert(database.Job).values(
        status=StatusEnum.queued, urls=urls,
    ).returning(database.Job.id)
    result = _execute(statement, commit)

    job_id: UUID = result.scalar()
    return job_id


def insert_logs(
    job_id: UUID, logs: Iterable[Log], *, commit: bool = True
) -> None:
    """Insert multiple log rows

    If `commit` is `True`, the statement will be run within a new
    transaction, which is then commited. Set this to `False` if you want
    to manage your own transactions.
    """
    # Convert the `utils` Logs to `database` Logs
    db_logs: list[database.Log] = []
    for log in logs:
        db_log = database.Log(
            job_id=job_id, level=log.level, message=log.message,
            timestamp=log.timestamp
        )
        db_logs.append(db_log)

    _add_all(db_logs, commit)


def _dedup_progress(
    progress_list: Sequence[Progress]
) -> Generator[Progress, None, None]:
    """Filter the given Progress list by only returning a single entry
    for each filename

    The last Progress entry in the list is yielded for each filename.

    The Progress entries are yielded in reverse order.

    This is needed because `on_conflict_do_update()` will raise an
    exception if you give it multiple rows that have the same conflict.
    """
    yielded_filenames = set()
    for progress in reversed(progress_list):
        if progress.filename not in yielded_filenames:
            yield progress
            yielded_filenames.add(progress.filename)


def upsert_progress(
    job_id: UUID, progress_list: Sequence[Progress], *, commit: bool = True
) -> None:
    """Update or insert multiple progress rows

    If `commit` is `True`, the statement will be run within a new
    transaction, which is then commited. Set this to `False` if you want
    to manage your own transactions.
    """
    # Convert the `utils` Progress entries to dicts that we can pass to
    # the statement
    db_progress_list: list[dict[str, Any]] = []
    for progress in _dedup_progress(progress_list):
        db_progress = {
            'job_id': job_id,
            'filename': progress.filename,
            'progress': progress.progress,
            'timestamp': progress.timestamp
        }
        db_progress_list.append(db_progress)

    # Postgres dialect is used because it supports
    # `on_conflict_do_update()`
    #
    # The values are baked into the statement instead of being passed to
    # `execute()` because SQLAlchemy will execute a separate statement
    # for each row otherwise when using `on_conflict_do_update()`
    statement = postgres_insert(database.Progress)
    statement = statement.on_conflict_do_update(
        # Replace existing row if a row already exists for the file
        # that's currently downloading (same job_id and filename)
        index_elements=(database.Progress.job_id, database.Progress.filename),
        set_={
            database.Progress.progress: statement.excluded.progress,
            database.Progress.timestamp: statement.excluded.timestamp,
        }
    ).values(db_progress_list)
    _execute(statement, commit)



"""Load the ytdl-server configuration from a config file and env-vars"""

from __future__ import annotations

__all__ = ('Config',)

import logging
import os
from typing import TYPE_CHECKING

import yaml

from .config_option import CELERY_OPTIONS, FLASK_OPTIONS, OPTIONS

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable
    from typing import Any, Final, Mapping, Optional

    from .config_option import FlaskCeleryOption, Option


# TODO: Add unit tests for Config
class Config:
    """Class that loads application config settings from env vars and
    from a config file"""

    __slots__ = (
        '_logger', 'database_uri', 'flask_config', 'celery_config',
        'download_dir', 'force_download_dir', 'ytdl_default_opts',
        'allow_user_opts', 'ytdl_enable_whitelist', 'ytdl_whitelist',
        'ytdl_whitelist_add', 'ytdl_whitelist_remove', 'ytdl_sensitive_opts',
        'custom_default_opts', 'custom_enable_whitelist', 'custom_whitelist',
        'ytdl_module', 'ytdl_class', 'daterange_module', 'daterange_class',
        'metadata_module', 'metadata_class'
    )

    database_uri: str
    """URI of the SQLAlchemy database"""
    flask_config: dict[str, Any]
    """Flask config options that are passed directly to Flask"""
    celery_config: dict[str, Any]
    """Celery config options that are passed directly to Celery"""
    download_dir: str
    """Dir that videos will be downloaded to by default"""
    force_download_dir: bool
    """Whether or not to force videos to be downloaded to `download_dir`"""
    allow_user_opts: bool
    """Whether or not to allow users to supply their own ytdl_opts and
    custom_opts"""
    ytdl_default_opts: dict[str, Any]
    """Default ytdl_opts"""
    ytdl_enable_whitelist: bool
    """Whether or not to use a whitelist for user-supplied ytdl_opts"""
    ytdl_whitelist: frozenset[str]
    """Set of whitelisted user ytdl_opts"""
    ytdl_whitelist_add: frozenset[str]
    """Additional ytdl_opts to add to `ytdl_whitelist`"""
    ytdl_whitelist_remove: frozenset[str]
    """Blacklisted ytdl_opts to remove from `ytdl_whitelist`"""
    ytdl_sensitive_opts: tuple[str, ...]
    """List of sensitive ytdl_opts whose values will be censored"""
    custom_default_opts: dict[str, Any]
    """Default custom_opts"""
    custom_enable_whitelist: bool
    """Whether or not to use a whitelist for custom_opts"""
    custom_whitelist: frozenset[str]
    """Set of whitelisted user custom_opts"""
    ytdl_module: str
    """The module that contains the YoutubeDL class"""
    ytdl_class: str
    """The YoutubeDL class to use"""
    daterange_module: str
    """The module that contains the DateRange class

    Defaults to the same value as `ytdl_module`
    """
    daterange_class: str
    """The DateRange class to use

    This is used for the 'datebefore' and 'dateafter' custom_opts
    """
    metadata_module: str
    """The module that contains the MetadataParser postprocessor class

    Defaults to the same value as `ytdl_module`
    """
    metadata_class: str
    """The MetadataParser postprocessor class to use

    This is used for the 'parse_metadata' and 'replace_in_metadata'
    custom_opts
    """

    _CONFIG_ENV_VAR: Final = 'YTDL_CONFIG'
    """Name of the env var that contains the filepath of the YAML config
    file that will be loaded (if it exists)"""

    _logger: Optional[logging.Logger]

    def __init__(
        self, *, logger: Optional[logging.Logger] = None,
        config_dict: Optional[Mapping[Any, Any]] = None,
        load_env_vars: bool = True
    ) -> None:
        """
        logger: Logger to log debug messages to. If `None`, no messages
            will be logged
        config_dict: Use this dict instead of loading the config file
        load_env_vars: Whether or not to load config values from their
            corresponding env vars

        `config_dict` and `load_env_vars` are intended to be used for
        testing
        """
        self._logger = logger
        if config_dict is None:
            config_dict = self._load_config_dict()

        for option in OPTIONS:
            value = self._get_value(option, config_dict, load_env_vars)
            setattr(self, option.name, value)

        if load_env_vars:
            self._load_flask_env_vars()
            self._load_celery_env_vars()

        self._run_manual_hacks()

    def _load_config_dict(self) -> dict[Any, Any]:
        """Load the YAML config file from the env var `_CONFIG_ENV_VAR`

        If the env var isn't set, returns an empty dict
        """
        if self._CONFIG_ENV_VAR in os.environ:
            config_path = os.environ[self._CONFIG_ENV_VAR]
            if config_path:
                self.log(
                    logging.INFO,
                    'Loading config file from env var: %s', config_path
                )
                with open(config_path, 'r') as file:
                    config_dict = yaml.safe_load(file)

                if not isinstance(config_dict, dict):
                    type_name = type(config_dict).__name__
                    raise ValueError(
                        'Invalid top-level type. '
                        f'Expected dict. Got {type_name}'
                    )
                return config_dict

        return {}

    def _get_value(
        self, option: Option, config_dict: Mapping[Any, Any],
        load_env_var: bool = True
    ) -> Any:
        """Get the value of the given option from the config dict or
        the corresponding env var

        Env vars have priority over the config dict

        If the option doesn't exist in the dict or as an env var, the
        default value for the option be set if it has one. Otherwise,
        an error will be raised

        If `load_env_var` is `False`, the env var won't be used
        """
        found_value = False
        is_default = False

        # Get the value from the env var if it exists
        if load_env_var and option.env_var is not None:
            if option.env_var in os.environ and os.environ[option.env_var]:
                value = option.get_env()
                found_value = True
                self.log(
                    logging.INFO, '%s: Found env var: %s. Value: %r',
                    option.name, option.env_var, value
                )

        if not found_value:
            if option.nested_in(config_dict, assert_type=True):
                # Get the value from the config file
                value = option.nested_get(config_dict)
                self._check_type(option, value)
                found_value = True

        if not found_value:
            if option.required:
                # Raise an error when the option is required, but wasn't
                # found
                raise KeyError(
                    f'Missing required config option: {option.name}'
                )
            else:
                value = option.get_default()
                is_default = True

        if not is_default:
            converted_value = option.convert_value(value)
            self.log(
                logging.INFO,
                '%s: Got value: %r', option.name, converted_value
            )
            return converted_value
        else:
            return value

    def _check_type(self, option: Option, value: Any) -> None:
        """Verify that the value of the option is the correct type"""
        if type(value) is not option.type:
            raise TypeError(
                f'The config option {option.name!r} has an invalid type. '
                f'Expected {option.type.__name__}. '
                f'Got {type(value).__name__}'
            )

        # Verify that all dicts only contain string keys
        #
        # Flask and Celery both only use string keys for their
        # configs
        if type(value) is dict:
            for key in value.keys():
                if type(key) is not str:
                    raise ValueError(
                        f"Invalid config option: '{option.name}.{key}'."
                        ' All keys must be strings'
                    )

    def _load_generic_env_vars(
        self, options: Iterable[FlaskCeleryOption], config_dict_name: str
    ) -> None:
        """Load all Flask and Celery env vars

        The env-var values are added to `flask_config` and
        `celery_config` respectively
        """
        config_dict: dict[str, Any] = getattr(self, config_dict_name)

        for option in options:
            if option.env_var in os.environ and os.environ[option.env_var]:
                value = option.get_env()
                self.log(
                    logging.INFO, '%s.%s: Found env var. Value: %r',
                    config_dict_name, option.name, value
                )
                config_dict[option.name] = value

    def _load_flask_env_vars(self) -> None:
        self._load_generic_env_vars(FLASK_OPTIONS, 'flask_config')

    def _load_celery_env_vars(self) -> None:
        self._load_generic_env_vars(CELERY_OPTIONS, 'celery_config')

    def _run_manual_hacks(self) -> None:
        """Option-specific hacks that can't be done automatically via
        `_Option`"""
        # Have Flask-SQLAlchemy use the database_uri
        self.flask_config['SQLALCHEMY_DATABASE_URI'] = self.database_uri

        # Have Celery use the SQL database as the result-backend by
        # default
        if 'result_backend' not in self.celery_config:
            result_backend = f'db+{self.database_uri}'
            self.log(
                logging.DEBUG,
                "Automatically setting `celery_config.result_backend` to %r",
                result_backend
            )
            self.celery_config['result_backend'] = result_backend

        # Have `daterange_module` and `metadata_module` use the same
        # module as `ytdl_module` by default
        for module in ('daterange_module', 'metadata_module'):
            if getattr(self, module) is None:
                self.log(
                    logging.DEBUG,
                    "Automatically setting `%s` to %r",
                    module, self.ytdl_module
                )
                setattr(self, module, self.ytdl_module)

    def items(self) -> Generator[tuple[str, Any], None, None]:
        """Return all config options as key-value pairs"""
        for option in OPTIONS:
            value = getattr(self, option.name)
            yield option.name, value

    def __repr__(self) -> str:
        options = []
        for option, value in self.items():
            options.append(f'{option}={value!r}')

        return f'Config({", ".join(options)})'

    def log(self, level: int, msg: str, *args: Any, **kwargs: Any) -> None:
        if self._logger is not None:
            self._logger.log(level, msg, *args, **kwargs)



"""Flask blueprint for getting the ytdl-server configuration"""

from __future__ import annotations

__all__ = ('bp', 'get_config')

from typing import TYPE_CHECKING

import flask

from .ytdl import censor_opts

if TYPE_CHECKING:
    from typing import Any, Final

    from .config import Config

bp = flask.Blueprint('config', __name__)

_WHITELISTED_OPTS: Final = frozenset((
    'download_dir', 'force_download_dir', 'allow_user_opts',
    'ytdl_enable_whitelist', 'ytdl_sensitive_opts', 'custom_default_opts',
    'custom_enable_whitelist'
))


def _create_config_dict(config: Config) -> dict[str, Any]:
    config_dict = {}
    for opt, value in config.items():
        if opt in _WHITELISTED_OPTS:
            config_dict[opt] = value

    # Censor 'ytdl_default_opts' so that it doesn't contain passwords
    config_dict['ytdl_default_opts'] = censor_opts(
        config.ytdl_default_opts, sensitive_opts=config.ytdl_sensitive_opts
    )

    if config.allow_user_opts and config.ytdl_enable_whitelist:
        # Merge 'ytdl_whitelist' with 'ytdl_whitelist_add' and
        # 'ytdl_whitelist_remove' so that the user only sees the final
        # whitelist
        merged_whitelist = set(config.ytdl_whitelist)
        merged_whitelist |= config.ytdl_whitelist_add
        merged_whitelist -= config.ytdl_whitelist_remove
        config_dict['ytdl_whitelist'] = tuple(sorted(merged_whitelist))
    else:
        # Set 'ytdl_whitelist' to 'None' if it's not being used
        config_dict['ytdl_whitelist'] = None

    if config.allow_user_opts and config.custom_enable_whitelist:
        config_dict['custom_whitelist'] = tuple(
            sorted(config.custom_whitelist)
        )
    else:
        # Set 'custom_whitelist' to 'None' if it's not being used
        config_dict['custom_whitelist'] = None

    return config_dict


@bp.route('', methods=('GET',))
def get_config() -> tuple[dict[str, Any], int]:
    """Return the server ytdl configuration"""
    config = flask.current_app.config['YTDL_CONFIG']

    config_dict = _create_config_dict(config)
    return config_dict, 200



"""HTTP exceptions that are used as Flask error responses

This module is also available within the `error` module. Importing it
from there is preferred, eg:
    from ytdl_server.error import http_error
    print(http_error.DataBadType)
"""

from __future__ import annotations

__all__ = (
    'CustomOptionBadValue', 'DataBadType', 'DataMissingKey',
    'ForbiddenCustomOption', 'ForbiddenYTDLOption', 'JobAlreadyCompleted',
    'JobNotFound', 'JobNotRunning', 'KeyBadType', 'KeyBadValue',
    'UnspecifiedError', 'YTDLOptionBadValue', 'YTDLUserOptionsNotAllowed',
    'TYPE_NAMES', 'INDEFINITE_TYPE_NAMES'
)

from typing import TYPE_CHECKING
from uuid import UUID

from .error_base import YTDLServerError

if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Any, Final, Optional, Union

    from .type import JSONableTypes

TYPE_NAMES: Final = {
    dict: 'object',
    list: 'array',
    tuple: 'array',
    str: 'string',
    int: 'integer',
    float: 'float',
    bool: 'boolean',
    type(None): 'null'
}
"""Mapping of Python types to their corresponding names in JSON

Note that this makes a distinction between `int` and `float` even though
JSON technically only has a single 'number' type.
"""

INDEFINITE_TYPE_NAMES: Final = {
    dict: 'an object',
    list: 'an array',
    tuple: 'an array',
    str: 'a string',
    int: 'an integer',
    float: 'a float',
    bool: 'a boolean',
    type(None): 'null'
}
"""Same as `TYPE_NAMES`, but with 'a' or 'an' prepended to the name"""


# TODO: Add tests for these
class HTTPError(YTDLServerError):
    """Base class for errors raised by the Flask REST API

    These errors are returned to the user.

    NOTE: This class must not be raised directly; use a subclass
    instead.
    """

    __slots__ = ('code', 'description')

    code: int
    """HTTP status code"""
    description: str
    """Human-readable error message"""

    def __init__(self, code: int, description: str) -> None:
        self.code = code
        self.description = description

    def to_dict(self) -> dict[str, Any]:
        """Create a dict that can be used as the JSON response data

        All attributes defined in `__slots__` that do not start with an
        underscore will be added to the dict. This includes the slots of
        parent classes.

        Additionally, the name of the class will be added as the `error`
        key.
        """
        response_data = {}

        for attr in self._get_all_slots():
            # Skip private members
            if not attr.startswith('_'):
                value = getattr(self, attr)
                response_data[attr] = value

        error_name = type(self).__name__
        response_data['error'] = error_name

        return response_data

    @classmethod
    def _get_all_slots(cls) -> Generator[str, None, None]:
        """Recursively return all __slots__ of the class

        The class and all parent classes are checked in
        method-resolution order.

        Slots are only returned if the class the slot belongs to is a
        subclass of HTTPError.
        """
        for base_cls in cls.mro():
            if not issubclass(base_cls, HTTPError):
                continue

            if hasattr(base_cls, '__slots__'):
                for slot in base_cls.__slots__:
                    yield slot


class UnspecifiedError(HTTPError):
    """Generic fallback error used by Flask"""

    __slots__ = ('name',)

    name: str
    """Short name describing the error"""

    def __init__(
        self, name: str, code: Optional[int], description: Optional[str]
    ) -> None:
        """`code` and `description` are optional because Flask's
        `HTTPException` has them set as optional
        """
        if code is None:
            # Not sure why this is optional, so let's just raise 500
            code = 500
        if description is None:
            description = name
        super().__init__(code, description)
        self.name = name


class DataBadType(HTTPError):
    """The request data is missing or is the wrong type"""

    __slots__ = ('type',)

    type: str
    """Expected JSON type, eg 'object', 'array'"""

    def __init__(
        self, type_: JSONableTypes, *, description: Optional[str] = None
    ) -> None:
        """
        type_: Expected Python type, eg `dict`, `str`, `type(None)`
        """
        if description is None:
            type_name = INDEFINITE_TYPE_NAMES[type_]
            description = f'Request data must be {type_name}'

        super().__init__(400, description)
        self.type = TYPE_NAMES[type_]


class DataMissingKey(HTTPError):
    """The request data is missing a required top-level key"""

    __slots__ = ('key',)

    key: str
    """Name of the missing key"""

    def __init__(
        self, key: str, *, description: Optional[str] = None
    ) -> None:
        if description is None:
            description = f'Missing key in request data: {key}'
        super().__init__(400, description)
        self.key = key


class KeyBadType(HTTPError):
    """A key within the request data object has the wrong type"""

    __slots__ = ('key', 'type')

    key: str
    """Name of the key"""
    type: str
    """Expected JSON type, eg 'object', 'array'"""

    def __init__(
        self, key: str, type_: JSONableTypes, *,
        description: Optional[str] = None
    ) -> None:
        """
        type_: Expected Python type, eg `dict`, `str`, `type(None)`
        """
        if description is None:
            type_name = INDEFINITE_TYPE_NAMES[type_]
            description = f'The key {key!r} must be {type_name}'
        super().__init__(400, description)
        self.key = key
        self.type = TYPE_NAMES[type_]


class KeyBadValue(HTTPError):
    """A key within the request data object has an invalid value"""

    __slots__ = ('key',)

    key: str
    """Name of the key"""

    def __init__(
        self, key: str, *, description: Optional[str] = None
    ) -> None:
        if description is None:
            description = f'The key {key!r} has an invalid value'
        super().__init__(400, description)
        self.key = key


class YTDLUserOptionsNotAllowed(HTTPError):
    """The user tried to provide their own ytdl_opts or custom_opts
    when it's forbidden by the server config"""

    __slots__ = ('key',)

    key: str
    """Name of the key"""

    def __init__(
        self, key: str, *, description: Optional[str] = None
    ) -> None:
        if description is None:
            description = 'The server does not allow custom youtube-dl options'
        super().__init__(403, description)
        self.key = key


class ForbiddenYTDLOption(HTTPError):
    """The user provided a ytdl_opt that isn’t whitelisted"""

    __slots__ = ('option',)

    option: str
    """Name of the ytdl_opt"""

    def __init__(
        self, option: str, *, description: Optional[str] = None
    ) -> None:
        if description is None:
            description = f'Forbidden ytdl_opt: {option}'
        super().__init__(403, description)
        self.option = option


class YTDLOptionBadValue(HTTPError):
    """The user provided a ytdl_opt with an invalid value"""

    __slots__ = ('option',)

    option: str
    """Name of the ytdl_opt"""

    def __init__(
        self, option: str, code: int, *, description: Optional[str] = None
    ) -> None:
        if description is None:
            description = f'The ytdl_opt {option!r} has an invalid value'
        super().__init__(code, description)
        self.option = option


class ForbiddenCustomOption(HTTPError):
    """The user provided a custom_opt that isn’t whitelisted"""

    __slots__ = ('option',)

    option: str
    """Name of the custom_opt"""

    def __init__(
        self, option: str, *, description: Optional[str] = None
    ) -> None:
        if description is None:
            description = f'Forbidden custom_opt: {option}'
        super().__init__(403, description)
        self.option = option


class CustomOptionBadValue(HTTPError):
    """The user provided a custom_opt with an invalid value"""

    __slots__ = ('option',)

    option: str
    """Name of the custom_opt"""

    def __init__(
        self, option: str, code: int, *, description: Optional[str] = None
    ) -> None:
        if description is None:
            description = f'The custom_opt {option!r} has an invalid value'
        super().__init__(code, description)
        self.option = option


class JobNotFound(HTTPError):
    """No job exists that matches the given job ID"""

    __slots__ = ('id',)

    id: str
    """Job ID"""

    def __init__(
        self, id_: Union[str, UUID], *, description: Optional[str] = None
    ) -> None:
        if description is None:
            description = 'Job ID not found'
        super().__init__(404, description)

        # Convert the UUID to a str if needed so that it's JSONable
        if isinstance(id_, UUID):
            id_ = str(id_)
        self.id = id_


class JobAlreadyCompleted(HTTPError):
    """The user tried to cancel a job that has already completed"""

    __slots__ = ('id', 'status_url')

    id: str
    """Job ID"""
    status_url: str
    """URL to get the status of the job"""

    def __init__(
        self, id_: Union[str, UUID], status_url: str, *,
        description: Optional[str] = None
    ) -> None:
        if description is None:
            description = (
                'Unable to cancel the job because it already completed'
            )
        super().__init__(400, description)

        # Convert the UUID to a str if needed so that it's JSONable
        if isinstance(id_, UUID):
            id_ = str(id_)
        self.id = id_
        self.status_url = status_url


class JobNotRunning(HTTPError):
    """The user tried to cancel a running job, but the Celery task
    isn’t running"""

    __slots__ = ('id', 'status_url')

    id: str
    """Job ID"""
    status_url: str
    """URL to get the status of the job"""

    def __init__(
        self, id_: Union[str, UUID], status_url: str, *,
        description: Optional[str] = None
    ) -> None:
        if description is None:
            description = (
                'Unable to cancel the job because it doesn\'t appear to be '
                'running'
            )
        super().__init__(500, description)

        # Convert the UUID to a str if needed so that it's JSONable
        if isinstance(id_, UUID):
            id_ = str(id_)
        self.id = id_
        self.status_url = status_url



"""Flask blueprint and functions for managing jobs"""

from __future__ import annotations

__all__ = ('bp', 'create_job', 'get_status', 'cancel_job')

import pathlib
import re
from typing import cast, TYPE_CHECKING

import flask
import celery.states

from . import db_util, task
from .error import http_error, JobNotFoundError
from .util import has_path_seps, StatusEnum

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any, Union
    from uuid import UUID

bp = flask.Blueprint('job', __name__)


def _check_outtmpl_dir(outtmpl: Union[str, dict[str, str]]) -> None:
    """Verifies that 'outtmpl' isn't being used to escape the download
    dir

    It works by verifying that 'outtmpl' is a child path within the
    download dir

    Also verifies the type of 'outtmpl'

    'outtmpl' can be a dict when using yt-dlp. In this case, every path
    in the dict is checked

    This function is safe to use with 'outtmpl' because youtube-dl
    removes path separators before injecting values into the template
    """
    config = flask.current_app.config['YTDL_CONFIG']

    if not config.force_download_dir:
        # Only check `outtmpl` if the relevant config option is enabled
        return

    if isinstance(outtmpl, str):
        # Convert str values to a single-item dict so we can iterate
        # over it
        #
        # The value of the key is unimportant
        outtmpl = {
            '_': outtmpl
        }
    elif not isinstance(outtmpl, dict):
        raise http_error.YTDLOptionBadValue(
            'outtmpl', 400, description=(
                "The ytdl_opt 'outtmpl' must be either a string or "
                'an object of strings'
            )
        )

    download_dir = config.download_dir
    parent_dir = pathlib.Path(download_dir).resolve()

    for filepath in outtmpl.values():
        if not isinstance(filepath, str):
            raise http_error.YTDLOptionBadValue(
                'outtmpl', 400, description=(
                    "All values within the 'outtmpl' object must be strings. "
                    f'Invalid value: {filepath!r}'
                )
            )

        # Append the download dir to the child path only if the child
        # path is relative
        if pathlib.Path(filepath).is_absolute():
            child_path = pathlib.Path(filepath).resolve()
        else:
            child_path = pathlib.Path(download_dir, filepath).resolve()

        if (
            parent_dir == child_path or
            not child_path.is_relative_to(parent_dir)
        ):
            raise http_error.YTDLOptionBadValue(
                'outtmpl', 403, description=(
                    "The ytdl_opt 'outtmpl' must be either a relative path or "
                    'a subpath of the download directory. '
                    f'Download directory: {download_dir!r}. '
                    f'Invalid path: {filepath!r}'
                )
            )


def _check_path_seps(opt_name: str, opt_value: str) -> None:
    """Verifies that the given ytdl_opt doesn't contain path separators

    Used for the following ytdl_opts:
        'outtmpl_na_placeholder', 'merge_output_format', 'final_ext'

    Path seperators are removed because these options can escape the
    download dir if a value such as '../../..' is given

    Unlike 'outtmpl', youtube-dl does not remove path separators when
    injecting variables into the template, so we can't use the parent
    method like in `_check_outtmpl_dir()`
    """
    config = flask.current_app.config['YTDL_CONFIG']
    if not config.force_download_dir:
        # Only check for separators if the relevant config option is
        # enabled
        return

    if not isinstance(opt_value, str):
        raise http_error.YTDLOptionBadValue(
            opt_name, 400,
            description=f'The ytdl_opt {opt_name!r} must be a string'
        )

    if has_path_seps(opt_value):
        raise http_error.YTDLOptionBadValue(
            opt_name, 403, description=(
                f'The ytdl_opt {opt_name!r} must not contain path separators'
            )
        )


def _create_job_check_input() -> None:
    """Verify that the user input received by `create_job()` is valid

    If the input is valid, it returns `None`

    If the input isn't valid, it raises `http_error.HTTPError`, which
    should be returned to the user as the response data.
    """
    # Assert that the content type is JSON
    if not flask.request.is_json:
        raise http_error.DataBadType(
            dict, description='Content-Type must be \'application/json\''
        )

    # Assert that the request data is the right type
    if not isinstance(flask.request.json, dict):
        raise http_error.DataBadType(dict)

    # Assert that 'urls' exists in the request data
    if 'urls' not in flask.request.json:
        raise http_error.DataMissingKey('urls')

    urls = flask.request.json['urls']

    # Assert that 'urls' is a non-empty list
    if not isinstance(urls, list):
        raise http_error.KeyBadType('urls', list)
    if not urls:
        raise http_error.KeyBadValue(
            'urls', description="The 'urls' array must not be empty"
        )

    # Check user-supplied ytdl_opts
    ytdl_user_opts = flask.request.json.get('ytdl_opts', None)
    if ytdl_user_opts is not None:
        _check_ytdl_user_opts(ytdl_user_opts)

    # Check custom_opts
    custom_opts = flask.request.json.get('custom_opts', None)
    if custom_opts is not None:
        _check_custom_opts(custom_opts)


def _check_ytdl_user_opts(ytdl_user_opts: Any) -> None:
    """Verify that the user-supplied ytdl_opts are valid

    Raises `http_error.HTTPError` if the ytdl_opts are invalid
    """
    if not isinstance(ytdl_user_opts, dict):
        raise http_error.KeyBadType('ytdl_opts', dict)
    if not ytdl_user_opts:
        # Skip checks if `ytdl_user_opts` is empty
        return None

    config = flask.current_app.config['YTDL_CONFIG']
    logger = flask.current_app.logger

    if not config.allow_user_opts:
        raise http_error.YTDLUserOptionsNotAllowed('ytdl_opts')

    if config.ytdl_enable_whitelist:
        allowed_opts = set(config.ytdl_whitelist)
        allowed_opts |= config.ytdl_whitelist_add
        allowed_opts -= config.ytdl_whitelist_remove

        for opt in ytdl_user_opts:
            if opt not in allowed_opts:
                raise http_error.ForbiddenYTDLOption(opt)

    # Check the value of individual ytdl_opts
    try:
        if 'outtmpl' in ytdl_user_opts:
            _check_outtmpl_dir(ytdl_user_opts['outtmpl'])

        for opt in (
            'outtmpl_na_placeholder', 'merge_output_format', 'final_ext'
        ):
            if opt in ytdl_user_opts:
                _check_path_seps(opt, ytdl_user_opts[opt])
    except http_error.YTDLOptionBadValue as e:
        # Log the error before re-raising it
        logger.debug(
            'create_job: %s: Invalid user-supplied ytdl_opt. Aborting',
            e.option
        )
        raise


def _check_custom_opt_extractaudio(custom_opts: Mapping[str, Any]) -> None:
    """Check the extractaudio custom_opt"""
    extractaudio = custom_opts.get('extractaudio', None)
    if extractaudio is None:
        return

    if not isinstance(extractaudio, dict):
        raise http_error.CustomOptionBadValue(
            'extractaudio', 400,
            description='Option must be an object or null'
        )

    audio_opts = (
        ('audioformat', str), ('audioquality', int),
        ('nopostoverwrites', bool)
    )
    for field, type_ in audio_opts:
        if field in extractaudio:
            value = extractaudio[field]
            if not isinstance(value, type_):
                type_name = http_error.INDEFINITE_TYPE_NAMES[type_]
                raise http_error.CustomOptionBadValue(
                    'extractaudio', 400,
                    description=f'The field {field!r} must be {type_name}'
                )

    if 'audioquality' in extractaudio and extractaudio['audioquality'] < 0:
        raise http_error.CustomOptionBadValue(
            'extractaudio', 400,
            description="The field 'audioquality' cannot be less than zero"
        )


def _check_custom_opt_split_chapters(custom_opts: Mapping[str, Any]) -> None:
    """Check the split_chapters custom_opt"""
    split_chapters = custom_opts.get('split_chapters', None)
    if split_chapters is None:
        return

    if not isinstance(split_chapters, dict):
        raise http_error.CustomOptionBadValue(
            'split_chapters', 400,
            description='Option must be an object or null'
        )

    force_keyframes = split_chapters.get('force_keyframes', None)
    if (
        force_keyframes is not None and
        not isinstance(force_keyframes, bool)
    ):
        if not isinstance(force_keyframes, bool):
            raise http_error.CustomOptionBadValue(
                'split_chapters', 400, description=(
                    "The field 'force_keyframes' must be a boolean or null"
                )
            )


def _check_custom_opt_remove_chapters(custom_opts: Mapping[str, Any]) -> None:
    """Check the remove_chapters custom_opt"""
    remove_chapters = custom_opts.get('remove_chapters', None)
    if remove_chapters is None:
        return

    if not isinstance(remove_chapters, dict):
        raise http_error.CustomOptionBadValue(
            'remove_chapters', 400,
            description='Option must be an object or null'
        )

    patterns = remove_chapters.get('patterns', None)
    ranges = remove_chapters.get('ranges', None)
    if not patterns and not ranges:
        raise http_error.CustomOptionBadValue(
            'remove_chapters', 400, description=(
                "The 'patterns' and/or 'ranges' fields must be non-empty "
                'arrays'
            )
        )

    if patterns is not None and not isinstance(patterns, list):
        raise http_error.CustomOptionBadValue(
            'remove_chapters', 400,
            description="The field 'patterns' must be an array or null"
        )
    if ranges is not None and not isinstance(ranges, list):
        raise http_error.CustomOptionBadValue(
            'remove_chapters', 400,
            description="The field 'ranges' must be an array or null"
        )

    if patterns:
        for pattern in patterns:
            if not isinstance(pattern, str):
                raise http_error.CustomOptionBadValue(
                    'remove_chapters', 400, description=(
                        "The field 'patterns' must only contain strings"
                    )
                )

            try:
                re.compile(pattern)
            except re.error:
                raise http_error.CustomOptionBadValue(
                    'remove_chapters', 400, description=(
                        "The field 'patterns' contains an invalid regular "
                        f'expression: {pattern}'
                    )
                )

    if ranges:
        for range_ in ranges:
            if not isinstance(range_, list):
                raise http_error.CustomOptionBadValue(
                    'remove_chapters', 400, description=(
                        "The field 'ranges' must only contain number-pair "
                        'arrays'
                    )
                )

            length = len(range_)
            if length != 2:
                raise http_error.CustomOptionBadValue(
                    'remove_chapters', 400, description=(
                        "The field 'ranges' contains an array with an "
                        f'invalid length. Expected 2. Got {length}'
                    )
                )

            for number in range_:
                if not isinstance(number, (int, float)):
                    raise http_error.CustomOptionBadValue(
                        'remove_chapters', 400, description=(
                            "The field 'ranges' contains an array with an "
                            'invalid type. Expected an array of numbers, '
                            f'not {range_}'
                        )
                    )

                if number < 0:
                    raise http_error.CustomOptionBadValue(
                        'remove_chapters', 400, description=(
                            "The field 'ranges' contains an array with a "
                            f'negative number: {range_}. Numbers must be '
                            'positive or zero'
                        )
                    )

            if range_[0] >= range_[1]:
                raise http_error.CustomOptionBadValue(
                    'remove_chapters', 400, description=(
                        "The field 'ranges' contains an array where the "
                        f'start is greater than the end: {range_}'
                    )
                )

    force_keyframes = remove_chapters.get('force_keyframes', None)
    if (
        force_keyframes is not None and
        not isinstance(force_keyframes, bool)
    ):
        raise http_error.CustomOptionBadValue(
            'remove_chapters', 400, description=(
                "The field 'force_keyframes' must be a boolean or null"
            )
        )


def _check_custom_opt_parse_metadata(custom_opts: Mapping[str, Any]) -> None:
    """Check the parse_metadata custom_opt"""
    parse_metadata = custom_opts.get('parse_metadata', None)
    if parse_metadata is None:
        return

    if not isinstance(parse_metadata, list):
        raise http_error.CustomOptionBadValue(
            'parse_metadata', 400,
            description='Option must be an array or null'
        )

    for interpreter in parse_metadata:
        if not isinstance(interpreter, dict):
            raise http_error.CustomOptionBadValue(
                'parse_metadata', 400,
                description='Option must only contain objects'
            )

        for field in ('from', 'to'):
            if field not in interpreter:
                raise http_error.CustomOptionBadValue(
                    'parse_metadata', 400,
                    description=f'Missing field: {field}'
                )

            value = interpreter[field]
            if not isinstance(value, str):
                raise http_error.CustomOptionBadValue(
                    'parse_metadata', 400,
                    description=f'The field {field!r} must be a string'
                )


def _check_custom_opt_replace_in_metadata(
    custom_opts: Mapping[str, Any]
) -> None:
    """Check the replace_in_metadata custom_opt"""
    replace_in_metadata = custom_opts.get('replace_in_metadata', None)
    if replace_in_metadata is None:
        return

    if not isinstance(replace_in_metadata, list):
        raise http_error.CustomOptionBadValue(
            'replace_in_metadata', 400,
            description='Option must be an array or null'
        )

    for replacer in replace_in_metadata:
        if not isinstance(replacer, dict):
            raise http_error.CustomOptionBadValue(
                'replace_in_metadata', 400,
                description='Option must only contain objects'
            )

        fields = (
            ('fields', list),
            ('regex', str),
            ('replace', str)
        )
        for field, type_ in fields:
            if field not in replacer:
                raise http_error.CustomOptionBadValue(
                    'replace_in_metadata', 400,
                    description=f'Missing field: {field}'
                )

            value = replacer[field]
            if not isinstance(value, type_):
                type_name = http_error.INDEFINITE_TYPE_NAMES[type_]
                raise http_error.CustomOptionBadValue(
                    'replace_in_metadata', 400,
                    description=f'The field {field!r} must be {type_name}'
                )

        for field in replacer['fields']:
            if not isinstance(field, str):
                raise http_error.CustomOptionBadValue(
                    'replace_in_metadata', 400,
                    description=(
                        "The field 'fields' must be an array of strings. "
                        f'Unexpected value: {field!r}'
                    )
                )


def _check_custom_opt_sponsorblock(custom_opts: Mapping[str, Any]) -> None:
    """Check the sponsorblock custom_opt"""
    sponsorblock = custom_opts.get('sponsorblock', None)
    if sponsorblock is None:
        return

    if not isinstance(sponsorblock, dict):
        raise http_error.CustomOptionBadValue(
            'sponsorblock', 400,
            description='Option must be an object or null'
        )

    mark = sponsorblock.get('mark', None)
    remove = sponsorblock.get('remove', None)

    if not mark and not remove:
        raise http_error.CustomOptionBadValue(
            'sponsorblock', 400, description=(
                "The 'mark' and/or 'remove' fields must be non-empty arrays"
            )
        )

    if mark is not None and not isinstance(mark, list):
        raise http_error.CustomOptionBadValue(
            'sponsorblock', 400,
            description="The field 'mark' must be an array or null"
        )
    if remove is not None and not isinstance(remove, list):
        raise http_error.CustomOptionBadValue(
            'sponsorblock', 400,
            description="The field 'remove' must be an array or null"
        )

    if mark:
        for category in mark:
            if not isinstance(category, str):
                raise http_error.CustomOptionBadValue(
                    'sponsorblock', 400, description=(
                        "The field 'mark' must only contain strings"
                    )
                )
    if remove:
        for category in remove:
            if not isinstance(category, str):
                raise http_error.CustomOptionBadValue(
                    'sponsorblock', 400, description=(
                        "The field 'remove' must only contain strings"
                    )
                )

    for field in ('template', 'api'):
        value = sponsorblock.get(field, None)
        if value is not None and not isinstance(value, str):
            raise http_error.CustomOptionBadValue(
                'sponsorblock', 400,
                description=f'The field {field} must be a string or null'
            )

    force_keyframes = sponsorblock.get('force_keyframes', None)
    if (
        force_keyframes is not None and
        not isinstance(force_keyframes, bool)
    ):
        if not isinstance(force_keyframes, bool):
            raise http_error.CustomOptionBadValue(
                'sponsorblock', 400, description=(
                    "The field 'force_keyframes' must be a boolean or null"
                )
            )


def _check_custom_opts(custom_opts: Any) -> None:
    """Verify that the user-supplied custom_opts are valid

    Raises `http_error.HTTPError` if the custom_opts are invalid
    """
    if not isinstance(custom_opts, dict):
        raise http_error.KeyBadType('custom_opts', dict)
    if not custom_opts:
        # Skip checks if `custom_opts` is empty
        return None

    config = flask.current_app.config['YTDL_CONFIG']
    logger = flask.current_app.logger

    # Verify that user-supplied custom_opts is allowed
    if not config.allow_user_opts:
        raise http_error.YTDLUserOptionsNotAllowed('custom_opts')

    # Verify that there are no non-whitelisted custom_opts
    if config.custom_enable_whitelist:
        for opt in custom_opts:
            if opt not in config.custom_whitelist:
                raise http_error.ForbiddenCustomOption(opt)

    # Check custom_opts that are supposed to be a str
    for opt in (
        'dateafter', 'datebefore', 'metafromtitle', 'remuxvideo',
        'recodevideo', 'convertsubtitles', 'convertthumbnails'
    ):
        value = custom_opts.get(opt, None)
        if value is not None and not isinstance(value, str):
            raise http_error.CustomOptionBadValue(
                opt, 400, description='Option must be a string or null'
            )

    # Check custom_opts that are supposed to be bools
    for opt in (
        'addmetadata', 'addchapters', 'embedsubtitles', 'embedthumbnail',
        'xattrs'
    ):
        if opt in custom_opts:
            value = custom_opts.get(opt, None)
            if value is not None and not isinstance(value, bool):
                raise http_error.CustomOptionBadValue(
                    opt, 400, description='Option must be a bool or null'
                )

    # Verify that vulnerable custom_opts don't contain path separators
    if config.force_download_dir:
        for opt in (
            'remuxvideo', 'recodevideo', 'convertsubtitles',
            'convertthumbnails'
        ):
            value = custom_opts.get(opt, None)
            if value is not None and has_path_seps(value):
                raise http_error.CustomOptionBadValue(
                    opt, 403,
                    description='Option must not contain path separators'
                )

    # Check individual custom_opts
    try:
        _check_custom_opt_extractaudio(custom_opts)
        _check_custom_opt_split_chapters(custom_opts)
        _check_custom_opt_remove_chapters(custom_opts)
        _check_custom_opt_parse_metadata(custom_opts)
        _check_custom_opt_replace_in_metadata(custom_opts)
        _check_custom_opt_sponsorblock(custom_opts)
    except http_error.CustomOptionBadValue as e:
        # Log the error before re-raising it
        logger.debug(
            'create_job: %s: Invalid user-supplied custom_opt. Aborting',
            e.option
        )
        raise


def _cancel_job_check_input() -> None:
    """Verify that the user input received by `cancel_job()` is valid

    If the input isn't valid, it raises `http_error.HTTPError`, which
    should be returned to the user as the response data.
    """
    # Assert that the content type is JSON
    if not flask.request.is_json:
        raise http_error.DataBadType(
            dict, description='Content-Type must be \'application/json\''
        )

    # Assert that the request data is the right type
    if not isinstance(flask.request.json, dict):
        raise http_error.DataBadType(dict)

    # Assert that 'status' exists in the request data
    if 'status' not in flask.request.json:
        raise http_error.DataMissingKey('status')

    status = flask.request.json['status']

    # Assert that the status is the proper value
    if status != StatusEnum.cancelled.name:
        raise http_error.KeyBadValue(
            'status', description=f'Unsupported status: {status}'
        )


@bp.route('/', methods=('POST',))
def create_job() -> tuple[dict[str, Any], int]:
    """Create a new job

    The request data must be a json object that matches the below format
        {
            "urls": [ "video1", "video2", "video3" ]
            "ytdl_opts": { ... }
            "custom_opts": { ... }
        }
    `ytdl_opts` is an optional object containing options that will be
    passed directly to youtube-dl

    `custom_opts` is an optional object containing custom ytdl options.
    See the documentation for more info

    `ytdl_opts` and `custom_opts` may be disabled or restricted
    depending on the server configuration
    """
    try:
        _create_job_check_input()
    except http_error.HTTPError as e:
        return e.to_dict(), e.code

    config = flask.current_app.config['YTDL_CONFIG']
    logger = flask.current_app.logger

    # Override type of the request json because it was verified in
    # `_create_job_check_input()`
    request_data = cast('dict[str, Any]', flask.request.json)
    urls = request_data['urls']
    ytdl_user_opts = request_data.get('ytdl_opts') or {}
    custom_user_opts = request_data.get('custom_opts') or {}

    logger.info('create_job: creating job')

    ytdl_opts = config.ytdl_default_opts | ytdl_user_opts
    custom_opts = config.custom_default_opts | custom_user_opts

    job_id = db_util.insert_job(urls=urls)
    logger.info('create_job: job_id: %s', job_id)

    logger.debug('create_job: %s: queuing task', job_id)
    task.download.apply_async(task_id=str(job_id), kwargs={
        'urls': urls,
        'ytdl_opts': ytdl_opts,
        'custom_opts': custom_opts,
        'sensitive_opts': config.ytdl_sensitive_opts,
        'download_dir': config.download_dir,
        'ytdl_module': config.ytdl_module,
        'ytdl_class': config.ytdl_class,
        'daterange_module': config.daterange_module,
        'daterange_class': config.daterange_class,
        'metadata_module': config.metadata_module,
        'metadata_class': config.metadata_class
    })

    # Return once the task is queued
    return {
        'id': job_id,
        'status_url': flask.url_for(
            '.get_status', job_id=job_id, _external=True
        )
    }, 202


@bp.route('/<uuid:job_id>', methods=('GET',))
def get_status(job_id: UUID) -> tuple[dict[str, Any], int]:
    """Get information about an existing job"""
    logger = flask.current_app.logger

    logger.debug('get_status: %s: Getting job status', job_id)
    try:
        job = db_util.get_job(job_id)
    except JobNotFoundError:
        logger.debug('get_status: %s: Job not found. Aborting', job_id)
        return_data = http_error.JobNotFound(job_id)
        return return_data.to_dict(), return_data.code

    logger.debug('get_status: %s: Getting logs', job_id)
    job.logs = db_util.get_logs(job_id)

    logger.debug('get_status: %s: Getting progress', job_id)
    job.progress = db_util.get_progress(job_id)

    return job.get_json(), 200


@bp.route('/<uuid:job_id>', methods=('PATCH',))
def cancel_job(job_id: UUID) -> tuple[dict[str, Any], int]:
    """Asynchronously cancel the given job by aborting the Celery task

    Does not wait for the job to be cancelled. The user must check the
    status of the job via `get_status()` if they want to know when it
    actually gets cancelled

    The request data must be a json object that matches the below format
        {
            "status": "cancelled"
        }
    """
    try:
        _cancel_job_check_input()
    except http_error.HTTPError as e:
        return e.to_dict(), e.code

    logger = flask.current_app.logger

    logger.debug('cancel_job: %s: Getting job status', job_id)
    try:
        status = db_util.get_job_status(job_id)
    except JobNotFoundError:
        logger.debug('cancel_job: %s: Job not found. exiting', job_id)
        return_data: http_error.HTTPError = http_error.JobNotFound(job_id)
        return return_data.to_dict(), return_data.code
    logger.debug('cancel_job: %s: Job status: %s', job_id, status.name)

    # Assert that the job is currently running
    if status in {StatusEnum.finished, StatusEnum.error, StatusEnum.timeout}:
        logger.debug('cancel_job: %s: Job already completed. Exiting', job_id)
        return_data = http_error.JobAlreadyCompleted(
            job_id, flask.url_for('.get_status', job_id=job_id, _external=True)
        )
        return return_data.to_dict(), return_data.code

    # Return if the job was already cancelled
    if status is StatusEnum.cancelled:
        logger.debug(
            'cancel_job: %s: Job is already cancelled. Exiting', job_id
        )
        return {
            'id': job_id,
            'status_url': flask.url_for(
                '.get_status', job_id=job_id, _external=True
            ),
            'description': (
                'The job is already cancelled. No action has been taken'
            )
        }, 200

    logger.debug('cancel_job: %s: Retrieving Celery task result', job_id)
    result = task.download.AsyncResult(str(job_id))
    logger.debug('cancel_job: %s: Task state: %r', job_id, result.state)

    # Return if the Celery task was already aborted
    #
    # Aborts are asynchronous, so the task doesn't always exit right
    # away, but the state updates immediately when `abort()` is called
    if result.state == 'ABORTED':
        logger.debug(
            'cancel_job: %s: Task was already aborted. Exiting', job_id
        )
        return {
            'id': job_id,
            'status_url': flask.url_for(
                '.get_status', job_id=job_id, _external=True
            ),
            'description': 'The job is already pending cancellation'
        }, 202

    # READY_STATES value: {'REVOKED', 'FAILURE', 'SUCCESS'}
    if result.state in celery.states.READY_STATES:
        logger.error(
            'cancel_job: %s: Task state: %r, Job status: %r: '
            'Celery task already completed, but the job status indicates that '
            'it should still be running. The task probably crashed '
            'unexpectedly',
            job_id, result.state, status.name
        )
        return_data = http_error.JobNotRunning(
            job_id, flask.url_for('.get_status', job_id=job_id, _external=True)
        )
        return return_data.to_dict(), return_data.code

    logger.info('cancel_job: %s: Aborting task', job_id)
    result.abort()

    return {
        'id': job_id,
        'status_url': flask.url_for(
            '.get_status', job_id=job_id, _external=True
        ),
        'description': 'The job is pending cancellation'
    }, 202



"""Class and helper functions for running youtube-dl"""

from __future__ import annotations

__all__ = ('Downloader', 'censor_opts')

from collections.abc import Sequence
from datetime import datetime, timezone
from typing import TYPE_CHECKING
import threading

from . import db_util
from .database import db
from .util import Log, LogEnum, Progress, StatusEnum

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable, Mapping
    from logging import Logger
    from typing import Any, Optional
    from uuid import UUID


class Downloader:
    """Class used by `task` to run YoutubeDL

    The class works by running YoutubeDL in a separate thread while
    handling all database transactions in the main thread. This prevents
    the download from slowing down due to slow database IO.

    `db_util.insert_job()` must be called before initializing this class
    so that the job exists in the database
    """

    __slots__ = (
        'job_id', 'urls', 'ytdl_opts', 'ytdl_class', 'logger', '_ytdl_logger',
        '_download_started', '_download_thread', '_progress_lock', '_log_lock',
        '_progress_queue', '_log_queue', '_download_exception'
    )

    job_id: UUID
    """ID of the job that's being downloaded"""
    urls: tuple[str, ...]
    """List of URLs to download"""
    ytdl_opts: dict[str, Any]
    """Dict of kwargs that will be passed to `ytdl_class`"""
    ytdl_class: type
    """YoutubeDL class to use

    Normally, this should be `youtube_dl.YoutubeDL` unless you're using
    a fork with a different module name
    """
    logger: Logger
    """Logger object to log to"""

    _ytdl_logger: YTDLLogger
    """YTDLLogger instance that's used for the job"""
    _download_started: bool
    """Whether or not `download()` has been called already"""
    _download_thread: threading.Thread
    """Thread that's used to run YoutubeDL"""
    _progress_lock: threading.Lock
    """Lock used to access `_progress_queue`"""
    _log_lock: threading.Lock
    """Lock used to access `_log_queue`"""
    _progress_queue: list[Progress]
    """List of progress entries that need to be written to the database"""
    _log_queue: list[Log]
    """List of logs that need to be written to the database"""
    _download_exception: Optional[Exception]
    """If the download thread fails, this will be set to the exception
    that was raised

    Otherwise, this will be set to `None` once the download thread exits
    normally.
    """

    def __init__(
        self, job_id: UUID, urls: Iterable[str], ytdl_opts: Mapping[str, Any],
        ytdl_class: type, logger: Logger
    ) -> None:
        self.job_id = job_id
        self.urls = tuple(urls)
        self.ytdl_opts = dict(ytdl_opts)
        self.ytdl_class = ytdl_class
        self.logger = logger

        self._download_started = False
        self._progress_lock = threading.Lock()
        self._log_lock = threading.Lock()
        self._progress_queue = []
        self._log_queue = []

        if 'progress_hooks' not in self.ytdl_opts:
            self.ytdl_opts['progress_hooks'] = (self.progress_hook,)
        else:
            self.ytdl_opts['progress_hooks'] = (
                tuple(ytdl_opts['progress_hooks']) + (self.progress_hook,)
            )

        # Prevent ytdl from injecting ANSI escape codes into the logs
        self.ytdl_opts['no_color'] = True
        # Prevent progress from being printed to the debug log (the
        # progress hook is used instead)
        self.ytdl_opts['noprogress'] = True
        # Disable youtube-dl version check
        # This is disabled by default. Set it explicitly so it can't be
        # enabled
        self.ytdl_opts['call_home'] = False

        self._ytdl_logger = self.YTDLLogger(self)
        self.ytdl_opts['logger'] = self._ytdl_logger

    def progress_hook(self, progress_map: dict[str, Any]) -> None:
        progress = Progress(progress_map, datetime.now(tz=timezone.utc))
        with self._progress_lock:
            self._progress_queue.append(progress)

    class YTDLLogger:
        __slots__ = ('_downloader',)

        _downloader: Downloader

        def __init__(self, downloader: Downloader) -> None:
            self._downloader = downloader

        def _insert(self, level: LogEnum, msg: str) -> None:
            log = Log(level, msg, datetime.now(tz=timezone.utc))
            with self._downloader._log_lock:
                self._downloader._log_queue.append(log)

        def debug(self, msg: str) -> None:
            self._insert(LogEnum.debug, msg)

        def warning(self, msg: str) -> None:
            self._insert(LogEnum.warning, msg)

        def error(self, msg: str) -> None:
            self._insert(LogEnum.error, msg)

    def _download(self) -> None:
        """Run youtube-dl

        This should be run in its own daemon thread
        """
        try:
            with self.ytdl_class(self.ytdl_opts) as ytdl:
                ytdl.download(self.urls)
        except Exception as e:
            self._download_exception = e
        else:
            self._download_exception = None

    def _set_final_status(self) -> None:
        """Set the status of the job after the download is complete"""
        if self._download_exception is None:
            self.logger.info('Job finished: %s', self.job_id)
            db_util.update_job(
                self.job_id,
                status=StatusEnum.finished, finished=db_util.CURRENT_TIMESTAMP
            )
        else:
            self.logger.error(
                'Job failed: %s', self.job_id,
                exc_info=self._download_exception
            )
            error_msg = (
                f'{type(self._download_exception).__name__}: '
                f'{self._download_exception}'
            )
            db_util.update_job(
                self.job_id,
                status=StatusEnum.error, finished=db_util.CURRENT_TIMESTAMP,
                # The error message is only logged from within `ytdl`
                # so that the user is only shown it when the error is
                # from youtube-dl
                error=error_msg
            )

    def _manage(self) -> None:
        """Manage the download from the main thread

        Performs the following tasks:
          - Every second, insert any logs and download progress that are
            generated by the download thread into the database.
          - Write the final job status to the database once the download
            thread completes.
        """
        download_is_running = True
        while download_is_running:
            self._download_thread.join(1)
            if not self._download_thread.is_alive():
                self.logger.debug('Download thread exited')
                download_is_running = False

            self._insert_logs_and_progress()

        self._set_final_status()

    def _insert_logs_and_progress(self) -> None:
        """Pop all logs and progress-entries from the queues, and insert
        them into the database

        The logs and progress-entries are inserted in a single
        transaction.
        """
        logs = self._pop_logs()
        progress = self._pop_progress()
        if not logs and not progress:
            return

        with db.session.begin():
            self._insert_logs(logs)
            self._insert_progress(progress)

    def _pop_logs(self) -> list[Log]:
        """Pop and return all logs from the queue"""
        with self._log_lock:
            logs = self._log_queue
            self._log_queue = []
        return logs

    def _pop_progress(self) -> list[Progress]:
        """Pop and return all progress-entries from the queue"""
        with self._progress_lock:
            progress = self._progress_queue
            self._progress_queue = []
        return progress

    def _insert_logs(self, popped_logs: Collection[Log]) -> None:
        """Insert the given logs into the database

        Does nothing if the collection is empty.

        This does not commit the SQL statement. Running it within a
        transaction is recommended.
        """
        if not popped_logs:
            return

        self.logger.debug('Inserting %d logs into DB', len(popped_logs))
        db_util.insert_logs(self.job_id, popped_logs, commit=False)

    def _insert_progress(self, popped_progress: Sequence[Progress]) -> None:
        """Insert the given progress entries into the database

        Does nothing if the sequence is empty.

        This does not commit the SQL statement. Running it within a
        transaction is recommended.
        """
        if not popped_progress:
            return

        self.logger.debug(
            'Inserting %d progress entries into DB', len(popped_progress)
        )
        db_util.upsert_progress(self.job_id, popped_progress, commit=False)

    def download(self) -> None:
        """Start the download

        This function can only be run once per instance. Attempting to
        run it again will raise `RuntimeError`
        """
        if self._download_started:
            raise RuntimeError('downloader can only be run once')
        self._download_started = True

        self.logger.info('Starting downloader for job %s', self.job_id)
        db_util.update_job(
            self.job_id,
            status=StatusEnum.downloading, started=db_util.CURRENT_TIMESTAMP
        )

        self._download_thread = threading.Thread(
            # Run as a daemon so that youtube-dl doesn't keep running
            # silently if the main thread crashes
            target=self._download, daemon=True
        )
        self._download_thread.start()

        self._manage()


def censor_opts(
    ytdl_opts: Mapping[str, Any], sensitive_opts: Iterable[str]
) -> dict[str, Any]:
    """Return a censored version of the given youtube-dl option dict
    that's safe to show to the user

    Sensitive options such as passwords are set to `None`

    ytdl_opts: Uncensored youtube-dl options
    sensitive_opts: List of keys that should be censored. If any of
        these options are in ytdl_opts, the value will be set to None
    """
    opts = dict(ytdl_opts)

    for opt in sensitive_opts:
        if opt in opts:
            opts[opt] = None

    return opts



"""Celery tasks"""

from __future__ import annotations

__all__ = ('download', 'TaskArgs')

import os
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING
from uuid import UUID

# billiard is used because it allows tasks to create child processes
from billiard import Process
from celery.exceptions import SoftTimeLimitExceeded
from celery.utils.log import get_task_logger

from . import db_util
from .celery import app
from .custom_opt import load_custom_opts, merge_custom_opts
from .database import db
from .util import import_object, StatusEnum, terminate
from .ytdl import Downloader, censor_opts

if TYPE_CHECKING:
    from logging import Logger
    from typing import Any, Iterable, Mapping, Sequence

    from .type import Terminatable

_logger = get_task_logger(__name__)


# Ignore typing error about the decorator being untyped
@app.task(bind=True)  # type: ignore[misc]
def download(
    self: Any, *, urls: Sequence[str], ytdl_opts: Mapping[str, Any],
    custom_opts: Mapping[str, Any], sensitive_opts: Iterable[str],
    download_dir: str, ytdl_module: str, ytdl_class: str,
    daterange_module: str, daterange_class: str, metadata_module: str,
    metadata_class: str
) -> None:
    """Download the given URLs via youtube-dl

    See `TaskArgs` for a documentation about the args

    `db_util.insert_job()` must be called first so that the job exists
    in the database
    """
    try:
        args = TaskArgs(
            urls=urls, ytdl_opts=ytdl_opts, custom_opts=custom_opts,
            sensitive_opts=sensitive_opts, download_dir=download_dir,
            ytdl_module=ytdl_module, ytdl_class=ytdl_class,
            daterange_module=daterange_module, daterange_class=daterange_class,
            metadata_module=metadata_module, metadata_class=metadata_class,
            logger=_logger
        )
        _real_download(self, args)
    except Exception:
        # Log unexpected errors, and set the job status to 'error'
        _logger.exception('Celery task failed')
        job_id = UUID(self.request.id)
        _set_job_status(job_id, StatusEnum.error)


@dataclass(frozen=True)
class TaskArgs:
    """Dataclass of arguments to pass to `_real_download()`

    This isn't used for `download()` because it's not JSONable
    """

    urls: Sequence[str]
    """List of URLs to download"""
    ytdl_opts: Mapping[str, Any]
    """ytdl_opts to pass to the YoutubeDL class

    custom_opts will be merged into this after they're parsed
    """
    custom_opts: Mapping[str, Any]
    """custom_opts that will be merged with ytdl_opts"""
    sensitive_opts: Iterable[str]
    """List of sensitive ytdl_opts, which will be censored before being
    added to the SQL database"""
    download_dir: str
    """Default download directory"""
    ytdl_module: str
    """Name of the module that ytdl_class will be imported from"""
    ytdl_class: str
    """Name of the YoutubeDL class"""
    daterange_module: str
    """Name of the module that daterange_class will be imported from"""
    daterange_class: str
    """Name of the DateRange class"""
    metadata_module: str
    """Name of the module that metadata_class will be imported from"""
    metadata_class: str
    """Name of the MetadataParserPP class"""

    logger: Logger
    """Logger to use for logging

    This is used by functions within `custom_opt` so that they don't
    depend on Celery. The `task` module uses the Celery logger directly.
    """


def _real_download(task: Any, args: TaskArgs) -> None:
    """Used by `download()` to download the URLs within a try-clause"""
    job_id = UUID(task.request.id)

    if task.is_aborted():
        _logger.info('Task aborted before the download started. Exiting')
        _set_job_status(job_id, StatusEnum.cancelled)
        return

    _logger.debug(
        'Importing youtube-dl class %r from module %r',
        args.ytdl_class, args.ytdl_module
    )
    YoutubeDL = import_object(args.ytdl_module, args.ytdl_class)

    _logger.debug('Changing to download dir: %s', args.download_dir)
    os.chdir(args.download_dir)

    _logger.debug('Loading custom_opts: %r', args.custom_opts)
    loaded_custom_opts = load_custom_opts(args)
    _logger.debug('Loaded custom_opts. Result: %r', loaded_custom_opts)

    merged_ytdl_opts = merge_custom_opts(args.ytdl_opts, loaded_custom_opts)

    censored_opts = censor_opts(merged_ytdl_opts, args.sensitive_opts)
    _logger.debug('Censored ytdl_opts: %r', censored_opts)
    db_util.update_job(job_id, ytdl_opts=censored_opts)

    # Run the downloader as a child process so that we can kill it if
    # the task is aborted or times out.
    downloader = Downloader(
        job_id=job_id, urls=args.urls,
        ytdl_opts=merged_ytdl_opts, ytdl_class=YoutubeDL,
        logger=_logger
    )
    process = Process(target=downloader.download)

    # Reset the DB connection pool so that the pool isn't shared with
    # the child process. Sharing the pool causes intermittent errors.
    #
    # More info:
    #   https://docs.sqlalchemy.org/en/14/core/pooling.html#using-connection-pools-with-multiprocessing-or-os-fork
    #
    # This must be called in the parent instead of the child because
    # calling it in the child doesn't completely eliminate the errors if
    # the parent accesses the pool after forking.
    db.engine.dispose()

    try:
        _logger.debug('Starting YTDL process')
        process.start()
        _logger.debug('Child PID: %s', process.pid)
        _ytdl_loop(job_id, process, task)
    # TODO: Add tests for this
    except SoftTimeLimitExceeded:
        # The time limit is configured via the `task_soft_time_limit`
        # Celery option
        _logger.warning('Task timed out. Terminating the YTDL process')
        terminate(process)
        _logger.debug('Process successfully terminated')
        _set_job_status(job_id, StatusEnum.timeout)


def _ytdl_loop(job_id: UUID, process: Terminatable, task: Any) -> None:
    """Continuously check the status of the YTDL process until it either
    exits or is aborted"""
    while True:
        # sleep() is used instead of join() because of this issue:
        # https://github.com/celery/billiard/issues/270
        time.sleep(1)
        # process.join(1)

        if not process.is_alive():
            _logger.info("YTDL process isn't running. Exiting")
            break

        if task.is_aborted():
            _logger.info('Task aborted. Terminating the YTDL process')
            terminate(process)
            _logger.debug('Process successfully terminated')
            _set_job_status(job_id, StatusEnum.cancelled)
            break


def _set_job_status(job_id: UUID, status: StatusEnum) -> None:
    """Set the status of the given job

    Also sets the 'finished' column to the current time
    """
    _logger.debug('Setting job status to %r', status.name)
    db_util.update_job(
        job_id, status=status, finished=db_util.CURRENT_TIMESTAMP
    )



"""Flask blueprint that contains a heath-check route"""

from __future__ import annotations

__all__ = ('bp', 'health_check')

from typing import TYPE_CHECKING

import flask

if TYPE_CHECKING:
    from typing import Any, Final

bp = flask.Blueprint('health', __name__)

_HEALTHY_RESPONSE: Final = {
    'healthy': True
}


@bp.route('', methods=('GET',))
def health_check() -> tuple[dict[str, Any], int]:
    """Health check

    Always returns `True` when the route is responding.
    """
    return _HEALTHY_RESPONSE, 200



"""Version numbers for ytdl-server"""

from __future__ import annotations

__all__ = ('MAIN', 'API', 'SCHEMA')

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Final

MAIN: Final = '1.1.7'
"""Main version number

This tracks all changes made to the program
"""

API: Final = '1.1.1'
"""REST API version

This only changes when there is a change to the REST API
"""

SCHEMA: Final = '1'
"""SQL schema version

This only changes when the schema of the SQL database changes
"""



"""Flask blueprint and utils for dealing with JSON"""

from __future__ import annotations

__all__ = (
    'bp', 'dumps_quiet', 'handle_exception', 'JSONEncoder',
    'JSONEncoderWithWarning'
)

from typing import cast, TYPE_CHECKING

import flask
import werkzeug

from .error import http_error

if TYPE_CHECKING:
    from typing import Any

bp = flask.Blueprint('json', __name__)


class JSONEncoder(flask.json.JSONEncoder):
    """JSONEncoder that converts unserializable objects to a string

    The object is lossily converted to a string representation.

    Doesn't require Flask app context
    """
    def _handle_bad_obj(self, obj: Any) -> Any:
        """Handle a bad object that isn't serializable

        You can change the behavior by overriding the function in a
        subclass
        """
        return f'{type(obj).__name__}: {obj}'

    def default(self, obj: Any) -> Any:
        try:
            # Attempt to convert the object using the additional methods
            # defined by flask.json.JSONEncoder
            return super().default(obj)
        except TypeError:
            return self._handle_bad_obj(obj)


class JSONEncoderWithWarning(JSONEncoder):
    """Subclass of JSONEncoder that additionally logs a warning to the
    current Flask app

    Requires Flask app context
    """
    def _handle_bad_obj(self, obj: Any) -> Any:
        flask.current_app.logger.warning(
            'Encountered unserializable object of type %r when dumping to '
            'json: %r',
            type(obj).__name__,
            obj
        )
        return super()._handle_bad_obj(obj)


def dumps_quiet(*args: Any, **kwargs: Any) -> str:
    """Wrapper function for `flask.json.dumps` that uses JSONEncoder"""
    kwargs['cls'] = JSONEncoder
    return flask.json.dumps(*args, **kwargs)


@bp.app_errorhandler(werkzeug.exceptions.HTTPException)
def handle_exception(e: werkzeug.exceptions.HTTPException) -> tuple[
    werkzeug.wrappers.Response, int
]:
    """Return JSON instead of HTML for HTTP errors"""
    # Start with the correct headers and status code from the error
    #
    # `get_response()` returns a `werkzeug.sansio.response.Response`
    # object, but `app_errorhandler()` expects a
    # `werkzeug.wrappers.Response` object.
    # This is based on an example from the Flask documentation, so it's
    # probably fine to cast it.
    # https://flask.palletsprojects.com/en/2.1.x/errorhandling/#generic-exception-handlers
    response = cast(werkzeug.wrappers.Response, e.get_response())

    # Create a JSONified response
    #
    # `flask.jsonify` is used instead of calling `flask.json.dumps`
    # directly because it pretty-prints the JSON depending on the app
    # config
    error = http_error.UnspecifiedError(e.name, e.code, e.description)
    jsonified_response = flask.jsonify(error.to_dict())

    # Replace the data in the original response with the JSONified data
    response.data = jsonified_response.data
    response.content_type = jsonified_response.content_type

    return response, error.code
